/*
 * [EncodingRecogniser.java]
 *
 * Summary: Help determine the encoding of a file by displaying it in various encodings.
 *
 * Copyright: (c) 2007-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2007-07-04 initial
 *  1.1 2008-04-01 display build number, increase size of sample
 *  1.2 2010-03-23 user selectable L&F. Correct bug in BOM recogniser.
 */
package com.mindprod.encodingrecogniser;

import com.mindprod.common18.Build;
import com.mindprod.common18.CMPAboutJBox;
import com.mindprod.common18.EIO;
import com.mindprod.common18.FontFactory;
import com.mindprod.common18.HybridJ;
import com.mindprod.common18.JEButton;
import com.mindprod.common18.Laf;
import com.mindprod.common18.VersionCheck;

import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.security.AccessControlException;
import java.util.Set;
import java.util.SortedMap;

/**
 * Help determine the encoding of a file by displaying it in various encodings.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.2 2010-03-23 user selectable L&F. Correct bug in BOM recogniser.
 * @since 2007-07-04
 */
@SuppressWarnings( { "FieldCanBeLocal" } )
public final class EncodingRecogniser extends JApplet
    {
    /**
     * height of Applet box in pixels. has no effect on application width. Does not include surrounding frame.  Height
     * to use in &lt;applet tag.
     *
     * @noinspection UnusedDeclaration
     */
    private static final int APPLET_HEIGHT = 528;

    /**
     * Width of Applet box in pixels, has no effect on application width. Height to use in &lt;applet tag.
     *
     * @noinspection UnusedDeclaration
     */
    private static final int APPLET_WIDTH = 636;

    private static final int FIRST_COPYRIGHT_YEAR = 2007;

    /**
     * maximum number of bytes in the file to read as a sample
     */
    private static final int MAX_BYTES_TO_SAMPLE = 1024 * 1024;

    /**
     * maximum number of bytes in the file to display in hex
     */
    private static final int MAX_HEX_BYTES_TO_SAMPLE = 22;

    /**
     * maximum number of chars in the file to display as hex
     */
    private static final int MAX_HEX_CHARS_TO_SAMPLE = 12;

    /**
     * non-displaying copyright.
     *
     * @noinspection UnusedDeclaration
     */
    private static final String EMBEDDED_COPYRIGHT =
            "Copyright: (c) 2007-2017 Roedy Green, Canadian Mind Products, http://mindprod.com";

    private static final String RELEASE_DATE = "2010-03-23";

    private static final String TITLE_STRING = "Encoding Recogniser";

    private static final String VERSION_STRING = "1.2";

    /**
     * background colour, pale green to match website
     */
    private static final Color BACKGROUND_FOR_BODY = Build.BACKGROUND_FOR_BLENDING;

    /**
     * background colour for instructions.  Default grey is too dark
     */
    private static final Color BACKGROUND_FOR_INSTRUCTIONS = new Color( 0xf8f8f8 );

    /**
     * instruction normal color
     */
    private static final Color FOREGROUND_FOR_INSTRUCTIONS = new Color( 0x339911 );

    /**
     * foreground colour for labels
     */
    private static final Color FOREGROUND_FOR_LABEL = new Color( 0x0000b0 );

    /**
     * for titles
     */
    private static final Color FOREGROUND_FOR_TITLE = new Color( 0xdc143c );

    /**
     * instructions font
     */
    private static final Font FONT_FOR_INSTRUCTIONS =
            FontFactory.build( "Dialog", Font.PLAIN, 12 );

    /**
     * title font
     */
    private static final Font FONT_FOR_TITLE = FontFactory.build( "Dialog", Font.BOLD, 15 );

    /**
     * for for title second line
     */
    private static final Font FONT_FOR_TITLE2 = FontFactory.build( "Dialog", Font.PLAIN, 14 );

    // table to convert a nibble to a hex char.
    private static final char[] hexChar = {
            '0',
            '1',
            '2',
            '3',
            '4',
            '5',
            '6',
            '7',
            '8',
            '9',
            'a',
            'b',
            'c',
            'd',
            'e',
            'f' };

    /**
     * choose a file to display
     */
    private JButton chooseFile;

    /**
     * chose which encoding to display
     */
    private JComboBox<String> chooseEncoding;

    /**
     * label for encoding
     */
    private JLabel encodingLabel;

    /**
     * label for file name
     */
    private JLabel fileLabel;

    /**
     * hex bytes label
     */
    private JLabel hexByteLabel;

    /**
     * hex char label
     */
    private JLabel hexCharLabel;

    /**
     * short instructions for use
     */
    private JLabel instructions;

    /**
     * Applet Title
     */
    private JLabel title;

    /**
     * title, second line
     */
    private JLabel title2;

    /**
     * Scroller fro displayFileContents.
     */
    private JScrollPane scroller;

    /**
     * file displayed as if it were encoding in the chosen encoding
     */
    private JTextArea displayFileContents;

    /**
     * name of file we are trying to figure out encoding for.
     */
    private JTextField fileName;

    /**
     * first few untranslated bytes of file in hex
     */
    private JTextField hexBytes;

    /**
     * first few translated chars of file in hex
     */
    private JTextField hexChars;

    /**
     * chars after bytes from file decoded
     */
    private String translatedBytes;

    /**
     * Raw sample from the file to use in display sample
     */
    private byte[] rawByteSample;

    /**
     * Fast convert a String to a hex string with possible leading zero.
     *
     * @param b      array of chars you want to convert to hex.
     * @param length how many chars you wanted convented to byte.
     *
     * @return hex equivalent of the bytes
     */
    private static String toHexString( String b, int length )
        {
        // StringBuilder is better than FastCat for char by char work.
        StringBuilder sb = new StringBuilder( length * 5 );
        for ( int i = 0; i < length; i++ )
            {
            if ( i != 0 )
                {
                sb.append( ' ' );
                }
            // work high order to low
            final int c = b.charAt( i );
            sb.append( hexChar[ ( c & 0xf000 ) >>> 12 ] );
            sb.append( hexChar[ ( c & 0xf00 ) >>> 8 ] );
            sb.append( hexChar[ ( c & 0xf0 ) >>> 4 ] );
            sb.append( hexChar[ c & 0x0f ] );
            }
        return sb.toString();
        }

    /**
     * Fast convert a byte array to a hex string with possible leading zero.
     *
     * @param b      array of bytes you want to convert to hex.
     * @param length how many bytes you wanted convented to byte.
     *
     * @return hex equivalent of the bytes
     */
    private static String toHexString( byte[] b, int length )
        {
        // StringBuilder is better than FastCat for char by char work.
        StringBuilder sb = new StringBuilder( length * 3 );
        for ( int i = 0; i < length; i++ )
            {
            if ( i != 0 )
                {
                sb.append( ' ' );
                }
            // look up high nibble char
            sb.append( hexChar[ ( b[ i ] & 0xf0 ) >>> 4 ] );
            // look up low nibble char
            sb.append( hexChar[ b[ i ] & 0x0f ] );
            }
        return sb.toString();
        }

    private void buildComponents()
        {
        title = new JLabel( TITLE_STRING + " " + VERSION_STRING );
        title.setFont( FONT_FOR_TITLE );
        title.setForeground( FOREGROUND_FOR_TITLE );
        title2 = new JLabel(
                "released:" +
                RELEASE_DATE +
                " build:" +
                Build.BUILD_NUMBER
        );
        title2.setFont( FONT_FOR_TITLE2 );
        title2.setForeground( FOREGROUND_FOR_TITLE );
        chooseFile = new JEButton( "Choose File" );
        chooseFile.setToolTipText( "Choose a file to study" );
        fileName = new JTextField();
        fileName.setEditable( false );
        fileName.setForeground( Color.RED );
        /* populate with all supported exceptions */
        SortedMap<String, Charset> sm = Charset.availableCharsets();
        Set<String> encodings = sm.keySet();
        chooseEncoding =
                new JComboBox<>( encodings.toArray( new String[ encodings.size() ] ) );
        try
            {
            // avoid looking at restricted system property.
            chooseEncoding.setSelectedItem( Charset.
                    defaultCharset().
                    name() );
            }
        catch ( AccessControlException e )
            {
            chooseEncoding.setSelectedIndex( 0 );
            }
        fileLabel = new JLabel( "file to check", JLabel.LEFT );
        fileLabel.setForeground( FOREGROUND_FOR_LABEL );
        encodingLabel = new JLabel( "encoding candidate", JLabel.RIGHT );
        encodingLabel.setForeground( FOREGROUND_FOR_LABEL );
        hexByteLabel = new JLabel( "raw hex bytes" );
        hexByteLabel.setForeground( FOREGROUND_FOR_LABEL );
        hexCharLabel = new JLabel( "decoded hex chars" );
        hexCharLabel.setForeground( FOREGROUND_FOR_LABEL );
        hexBytes = new JTextField();
        hexBytes.setEditable( false );
        hexChars = new JTextField();
        hexChars.setEditable( false );
        displayFileContents = new JTextArea();
        displayFileContents.setEditable( false );
        scroller =
                new JScrollPane( displayFileContents,
                        JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                        JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
        instructions =
                new JLabel(
                        "Choose a file, then try various encodings till you find an encoding that displays it " +
                        "properly.",
                        JLabel.CENTER
                );
        instructions.setFont( FONT_FOR_INSTRUCTIONS );
        instructions.setBackground( BACKGROUND_FOR_INSTRUCTIONS );
        instructions.setForeground( FOREGROUND_FOR_INSTRUCTIONS );
        }

    /**
     * build a menu with Look & Feel and About across the top
     */
    private void buildMenu()
        {
        // turn on anti-alias
        System.setProperty( "swing.aatext", "true" );
        final JMenuBar menubar = new JMenuBar();
        setJMenuBar( menubar );
        final JMenu lafMenu = Laf.buildLookAndFeelMenu();
        if ( lafMenu != null )
            {
            menubar.add( lafMenu );
            }
        final JMenu menuHelp = new JMenu( "Help" );
        menubar.add( menuHelp );
        final JMenuItem aboutItem = new JMenuItem( "About" );
        menuHelp.add( aboutItem );
        aboutItem.addActionListener( new ActionListener()
            {
            public void actionPerformed( ActionEvent e )
                {
                // open about frame
                new CMPAboutJBox( TITLE_STRING,
                        VERSION_STRING,
                        "Helps you determine a file's encoding.",
                        "",
                        "freeware",
                        RELEASE_DATE,
                        FIRST_COPYRIGHT_YEAR,
                        "Roedy Green",
                        "ENCODINGRECOGNISER",
                        "1.8" );
                }
            } );
        }

    /**
     * classify sample, looking for BOM
     */
    private void classifyBOM()
        {
        if ( rawByteSample.length < 4 )
            {
            return;
            }
        int b0 = rawByteSample[ 0 ] & 0xff;
        int b1 = rawByteSample[ 1 ] & 0xff;
        int b2 = rawByteSample[ 2 ] & 0xff;
        int b3 = rawByteSample[ 3 ] & 0xff;
        // ff fe 00 00	UTF-32 for 32-bit internal UCS-4, little endian, Intel/Microsoft order.
        // 00 00 fe ff	UTF-32 for 32-bit internal UCS-4, big-endian, Java network order
        // ff fe xx xx	UTF-16 for 16-bit internal UCS-2, little endian, Intel/Microsoft order
        // fe ff	UTF-16 for 16-bit internal UCS-2, big endian, Java network order
        // ef bb bf	UTF-8 endian, strictly speaking does not apply,
        // though it uses big-endian most-significant-bytes first representation.
        if ( b0 == 0xff && b1 == 0xfe && b2 == 0x00 && b3 == 0x00 )
            {
            instructions.setText( "Hint: the file starts with a UTF-32LE BOM." );
            }
        else if ( b0 == 0x00 && b1 == 0x00 && b2 == 0xfe && b3 == 0xff )
            {
            instructions.setText( "Hint: the file starts with a UTF-32BE BOM." );
            }
        else if ( b0 == 0xff && b1 == 0xfe )
            {
            instructions.setText( "Hint: the file starts with a UTF-16LE BOM." );
            }
        else if ( b0 == 0xfe && b1 == 0xff )
            {
            instructions.setText( "Hint: the file starts with a UTF-16BE BOM." );
            }
        else if ( b0 == 0xef && b1 == 0xbb && b2 == 0xbf )
            {
            instructions.setText( "Hint: the file starts with a UTF-8 BOM." );
            }
        else
            {
            instructions.setText( "Hint: the file has no Unicode BOM." );
            }
        }

    /**
     * display the sample raw bytes from the file with the current encoding
     */
    private void displaySampleWithCurrentEncoding()
        {
        String encodingName = ( String ) chooseEncoding.getSelectedItem();
        // Use String constructor to decode bytes to String.
        // We do this only once, so no need to bother with nio.
        try
            {
            int hexByteSampleSize =
                    Math.min( rawByteSample.length, MAX_HEX_BYTES_TO_SAMPLE );
            hexBytes.setText( toHexString( rawByteSample, hexByteSampleSize ) );
            translatedBytes = new String( rawByteSample, encodingName );
            int hexCharSampleSize =
                    Math.min( translatedBytes.length(), MAX_HEX_CHARS_TO_SAMPLE );
            hexChars.setText( toHexString( translatedBytes,
                    hexCharSampleSize ) );
            displayFileContents.setText( translatedBytes );
            }
        catch ( UnsupportedEncodingException e )
            {
            instructions.setText(
                    "Unsupported encoding.  Please report program bug." );
            }
        }

    /**
     * register listeners for various components.
     */
    private void hookComponents()
        {
        // what to do when user selects a new file.
        chooseFile.addActionListener( new ActionListener()
            {
            public void actionPerformed( ActionEvent e )
                {
                JFileChooser fc = new JFileChooser();
                // No filter. No suggested starting point.
                switch ( fc.showOpenDialog( EncodingRecogniser.this ) )
                    {
                    case JFileChooser.APPROVE_OPTION:
                        final File file = fc.getSelectedFile();
                        fileName.setText( EIO.getCanOrAbsPath( file ) );
                        // get text from the file.
                        try
                            {
                            int sampleSize = ( int ) Math.min( file.length(),
                                    MAX_BYTES_TO_SAMPLE );
                            // O P E N
                            FileInputStream fis = new FileInputStream( file );
                            // R E A D
                            byte[] b = new byte[ sampleSize ];
                            // -1 means eof.
                            // You don't necessarily get all you ask for in one read.
                            // You get what's immediately available.
                            int bytesRead = fis.read( b,
                                    0
                                    /* offset in ba */,
                                    sampleSize
                                    /* bytes to read */ );
                            if ( bytesRead != sampleSize )
                                {
                                throw new IOException( "cannot read file" );
                                }
                            rawByteSample = b;
                            // C L O S E
                            fis.close();
                            displaySampleWithCurrentEncoding();
                            classifyBOM();
                            }
                        catch ( IOException e2 )
                            {
                            fileName.setText( "Cannot read "
                                              + fileName.getText() );
                            }
                        break;
                    case JFileChooser.CANCEL_OPTION:
                    case JFileChooser.ERROR_OPTION:
                        break;
                    default:
                    }
                }
            } );
        chooseEncoding.addItemListener( new ItemListener()
            {
            /**
             * Called whenever the value of the encoding changes. W
             * @param e the event that characterizes the change.
             */
            public void itemStateChanged( ItemEvent e )
                {
                displaySampleWithCurrentEncoding();
                }
            } );
        }

    /**
     * layout the components
     *
     * @param contentPane contentPane of the JApplet
     */
    private void layoutComponents( Container contentPane )
        {
        // --0--------1-----------------2-
        // 0 title  ------------------
        // 1 title2 ------------------
        // 2 ----------"file"------"encoding"
        // 3 choose--  file-------- encoding
        // 4 "hex bytes"    --- hexBytes----
        // 5 "hex chars"    --- hexChars----
        // 6 ---file contents---------------
        // 7 ---- instructions --------------
        // x y w h wtx wty anchor fill T L B R padx pady
        contentPane.add( title,
                new GridBagConstraints( 0,
                        0,
                        2,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.WEST,
                        GridBagConstraints.NONE,
                        new Insets( 10, 10, 0, 10 ),
                        0,
                        0 )
        );
        contentPane.add( title2,
                new GridBagConstraints( 0,
                        1,
                        2,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.WEST,
                        GridBagConstraints.NONE,
                        new Insets( 0, 10, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( chooseFile,
                new GridBagConstraints( 0,
                        3,
                        1,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.WEST,
                        GridBagConstraints.NONE,
                        new Insets( 5, 10, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( fileLabel,
                new GridBagConstraints( 1,
                        2,
                        1,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.WEST,
                        GridBagConstraints.NONE,
                        new Insets( 5, 0, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( encodingLabel,
                new GridBagConstraints( 2,
                        2,
                        1,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.EAST,
                        GridBagConstraints.NONE,
                        new Insets( 5, 5, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( fileName,
                new GridBagConstraints( 1,
                        3,
                        1,
                        1,
                        50.0,
                        0.0,
                        GridBagConstraints.WEST,
                        GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 0, 5, 10 ),
                        250
                        /* allow room for new filename */,
                        0 )
        );
        contentPane.add( chooseEncoding,
                new GridBagConstraints( 2,
                        3,
                        1,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.EAST,
                        GridBagConstraints.NONE,
                        new Insets( 5, 10, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( hexByteLabel,
                new GridBagConstraints( 0,
                        4,
                        1,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.EAST,
                        GridBagConstraints.NONE,
                        new Insets( 5, 10, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( hexBytes,
                new GridBagConstraints( 1,
                        4,
                        2,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.EAST,
                        GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 0, 5, 10 ),
                        250,
                        0 )
        );
        contentPane.add( hexCharLabel,
                new GridBagConstraints( 0,
                        5,
                        1,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.EAST,
                        GridBagConstraints.NONE,
                        new Insets( 5, 10, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( hexChars,
                new GridBagConstraints( 1,
                        5,
                        2,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.EAST,
                        GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 0, 5, 10 ),
                        250,
                        0 )
        );
        contentPane.add( scroller,
                new GridBagConstraints( 0,
                        6,
                        3,
                        1,
                        100.0,
                        100.0,
                        GridBagConstraints.CENTER,
                        GridBagConstraints.BOTH,
                        new Insets( 5, 10, 5, 10 ),
                        0,
                        0 )
        );
        contentPane.add( instructions,
                new GridBagConstraints( 0,
                        7,
                        3,
                        1,
                        0.0,
                        0.0,
                        GridBagConstraints.CENTER,
                        GridBagConstraints.BOTH,
                        new Insets( 5, 10, 10, 10 ),
                        0,
                        0 )
        );
        }

    /**
     * Allow this Applet to run as as application as well.
     *
     * @param args command line arguments ignored.
     */
    public static void main( String args[] )
        {
        HybridJ.fireup( new EncodingRecogniser(),
                TITLE_STRING + " " + VERSION_STRING,
                APPLET_WIDTH,
                APPLET_HEIGHT );
        } // end main

    /**
     * Called by the browser or Applet viewer to inform
     * this Applet that it is being reclaimed and that it should destroy
     * any resources that it has allocated.
     */
    public void destroy()
        {
        chooseEncoding = null;
        chooseFile = null;
        displayFileContents = null;
        encodingLabel = null;
        fileLabel = null;
        fileName = null;
        hexByteLabel = null;
        hexBytes = null;
        hexCharLabel = null;
        hexChars = null;
        instructions = null;
        rawByteSample = null;
        scroller = null;
        title2 = null;
        title = null;
        translatedBytes = null;
        }

    /**
     * Called by the browser or Applet viewer to inform
     * this Applet that it has been loaded into the system.
     */
    @Override
    public void init()
        {
        // need signed to read file
        if ( !VersionCheck.isJavaVersionOK( 1, 8, 0, this ) )
            {
            // effectively abort
            return;
            }
        buildMenu(); // also initial L&F
        Container contentPane = this.getContentPane();
        contentPane.setLayout( new GridBagLayout() );
        contentPane.setBackground( BACKGROUND_FOR_BODY );
        buildComponents();
        layoutComponents( contentPane );
        hookComponents();
        validate();
        setVisible( true );
        }
    }
